
require 'rule'

$UNRES = Term.new("unRes")



class RewriterBody
  
  attr :nfn_body #formula in CNF
  attr :dlp_body #conjunction of literals
  attr :std_rules_chr #array with standard rule
  
  attr :Univ #array of predicate Univ
  attr :aux_hash #basic disjunction with corresponding atom
  attr :match #array match atoms
  attr :UnResSharedVars
  
  def initialize(b)
    @nfn_body = b
    @dlp_body = Conjunction.new([])
    @std_rules_chr = Array.new
    
    @Univ = Array.new
    @aux_hash = Hash.new
    @match = Array.new
    @UnResSharedVars = Terms.new([])
  end
  


  #################################################################
  # Private Mothods
  #################################################################

  #////////////////////////////////////////////////////////////////
  # Given a basic disjunction, function build_aux builds 
  # a new atom auxiliary  with fresh predicate name, 
  # representing the basic disjunction.
  #
  # @bd: basic disjunction
  # @n_rule: number of the NFN rule containing bd
  # 
  def build_aux(bd, n_rule, head)
    n_disj = bd.num_bd

    pred = "aux" + n_disj.to_s + "_" + n_rule.to_s
    
    terms = Terms.new([])
    safe_shared = (bd.safe_vars & @nfn_body.shared_vars) + 
      (bd.safe_vars & head.vars_disj)
    safe_shared = safe_shared.deleteDuplicates
    terms = safe_shared + bd.sharedVars + bd.unResVars
    
    return Atom.new(pred,terms)
  end
  
  #////////////////////////////////////////////////////////////////
  # Function atomsJoiningUnresVars
  # 
  # 1. renaming unrestricted variables appearing in each auxiliary 
  #    atom;
  # 2. build fresh atoms "match" to define an explicit match 
  #    between unrestricted variables and contant $UNRES;
  # 3. adds rules defining the predicate match.
  #
  # @nr: number rule
  def atomsJoiningUnresVars(nr)
    @nfn_body.unRes_vars.each do |sv|
      # predicate name for match atom corresponding to variable sv 
      pred = "match" + sv.term.to_s + "_" + nr.to_s
      svArray = Array.new #Array with new names of variable sv

      @aux_hash.each_key do |bd1|
	if bd1.unResVars.include?(sv)
	  # rename unrestricted shared variable sv  appearing in the
	  # corresponging auxiliary atom of bd1                
	  svNew = sv.term.to_s + bd1.num_bd.to_s + "_" + nr.to_s
	  svNewT = Term.new(svNew)

	  aux_hash[bd1].terms.set_term(aux_hash[bd1].terms.index(sv),svNewT)

	  svArray.push(svNew)
	end

      end


	tmp=svArray.length
 
	while tmp > 0
	  (tmp-1).times{|i|
	    if svArray[i] > svArray[i+1]
	      svArray[i] , svArray[i+1] = svArray[i+1] , svArray[i] #scambia svArray[i] con svArray[i+1]
	    end
	  }
	  tmp-=1
	end

      
      svArray
      # build atom match to define macching between X and $UNRES
      varCorr = svArray[0]
      for n in 1.. svArray.size - 1
	termsMatch = Terms.new([])           
	termsMatch.set_term(0, Term.new(varCorr))
	termsMatch.set_term(1, Term.new(svArray[n]))
	varCorr = sv.term.to_s + (n-1).to_s + n.to_s + "_" + nr.to_s
	termsMatch.set_term(2, Term.new(varCorr))
	@match = @match + [Atom.new(pred,termsMatch)]
      end
      bodyMatch = Atom.new("univ" + sv.term.to_s  + "_" +
			   nr.to_s,[sv])
      addUnivUnresRules(bodyMatch)
      
      # Add rule defining match predicate
      @std_rules_chr = @std_rules_chr +
	[Rule.new(Atom.new(pred,Terms.new([sv,sv,sv])),bodyMatch).to_s]
      @std_rules_chr = @std_rules_chr + 
	[Rule.new(Atom.new(pred,Terms.new([sv,$UNRES,sv])),bodyMatch).to_s]
      @std_rules_chr = @std_rules_chr + 
	[Rule.new(Atom.new(pred,Terms.new([$UNRES,sv,sv])),bodyMatch).to_s]
      @std_rules_chr = @std_rules_chr + 
	[Rule.new(Atom.new(pred,Terms.new([$UNRES,$UNRES,$UNRES])),[]).to_s]
     end
  end
    
  #////////////////////////////////////////////////////////////////
  # Function addUnivUnresRules build rules defining
  # univ predicate of unrestricted variables.
  # 
  # @univ: predicate to define
  def addUnivUnresRules(univ)
    nfn_body.each do |bd|
      bd.each do |lit|
	if  lit.isNegative == false &&
	    lit.internal_atom.terms.include?(univ.terms[0])
	  @std_rules_chr = @std_rules_chr + [Rule.new(univ, lit).to_s]
	end
      end
    end
  end  

  #////////////////////////////////////////////////////////////////
  # Function addUnivSharedRule build rules defining
  # univ predicate of safe variables.
  # 
  def addUnivSharedRules
    @Univ.each do |univ|
      body = Conjunction.new([])
      # atom univ contains only a variable
      var = univ.terms[0]
      # an aux atom is added to body only if var
      # is a safe variable of the corrisponding
      # basic disjunction

      @aux_hash.each_key  do |bd|
	num = bd.num_bd
	if bd.safe_vars.include?(var)
	  body = Conjunction.new([aux_hash[bd]])

	  @aux_hash.each_key  do |bd2|
	    if bd2.safe_vars.include?(var)
	      if bd2.num_bd < num 
		body = Conjunction.new([aux_hash[bd2]])
	      end
	      num = bd2.num_bd
	    end
	  end
	break
	end
      end
      @std_rules_chr = @std_rules_chr + [Rule.new(univ,body).to_s]
    end
  end
  
  #////////////////////////////////////////////////////////////////
  # Function atomsJoiningSharedVars
  # 
  # 1. it renames each shared variables X assuming value $UnRes 
  #    in all auxiliary atoms
  # 2. build fresh atoms "match" to define an explicit match 
  #    between shared variables and contant $UNRES;
  # 3. adds rules defining the predicate match.
  #
  # @nr: number rule
  def atomsJoiningSharedVars(nr)

    @aux_hash.each_key do |bd|
      nd = bd.num_bd
      bd.sharedVars.each do |sv|
	# rename each variable shared assuming special constant $UNRES
        # appearing in the corresponging auxiliary atom of bd
	if @UnResSharedVars.include?(sv)
	  svNew = sv.to_s + nd.to_s + "_" + nr.to_s
	  svNewT = Term.new(svNew)  

	  @aux_hash[bd].terms.set_term(aux_hash[bd].terms.index(sv),svNewT)
	  pred = "match" + sv.to_s  + "_" + nr.to_s
	  match_atom = Atom.new(pred, Terms.new([sv,svNew]))
          @match = @match + [match_atom]	  
	  
	  univ_atom = Atom.new("univ" + sv.term.to_s + "_" +  nr.to_s,[sv])
	  
	  # if univ_atom does not contained in @Univ
	  # then it is added to @Univ
          add = true

	  @Univ.each do |a|
	    if (a == univ_atom)
	      add = false
	    end
	  end
	  if add == true
	    @Univ = @Univ + [univ_atom]
	  end
	  
	  @std_rules_chr = @std_rules_chr + 
	    [Rule.new(Atom.new(pred,Terms.new([sv,sv])),univ_atom).to_s]
	  
	  @std_rules_chr = @std_rules_chr + 
	    [Rule.new(Atom.new(pred,Terms.new([sv,$UNRES])),univ_atom).to_s]
	end
      end   
    end
  end
  
  #////////////////////////////////////////////////////////////////
  # Given an auxiliary atom, function build_aux_rules
  # builds the rules that define the auxiliary predicate
  # The function substitutes constant $UNRES for each 
  # variable appearing only in the head of the rule.
  #
  # @a: auxiliary atom
  # @bd: basic disjunction
  # @nd: number disjunction
  
  def build_aux_rules(a,bd,nd,nr)
    
    # for each literal of bd a new rule is built
    bd.each do |l|
      head_aux_rule = a.copy
      body_aux_rule = Conjunction.new([l])
      
      # unbound is the array of variables appearing in
      # head_aux_rule and that do not appear in the
      # literal l 
      unbound = head_aux_rule.terms - l.internal_atom.terms
      
      # each variable of unbound in aux is substituted 
      # with the constant #unRes to build a safe rule
      unbound.each do |t|
	head_aux_rule.terms.set_term(head_aux_rule.terms.index(t), $UNRES)
	
	# Each safe shared variable assuming value $UNRES 
	# is added to @UnResSharedVars
	@UnResSharedVars = @UnResSharedVars.push(t)
      end
      
      # il l is a negative literal
      # for each shared variable of l a new atom
      # is added to body_aux_rule in order to build
      # a safe rule
      if l.isNegative == true
	l.internal_atom.terms.each do |sv|
	  pred = "univ" + sv.to_s + "_" + nr.to_s
	  ts = Terms.new([sv])
          univ_atom = Atom.new(pred,ts)
	  body_aux_rule.add(univ_atom)
	  
	  # if univ_atom does not contained in @Univ
	  # then it is added to @Univ
          add = true
	  @Univ.each do |a|
	    if a == univ_atom
	      add = false
	    end
	  end
	  if add == true

	    @Univ = @Univ + [univ_atom]
	  end
	end
      end
      @std_rules_chr = @std_rules_chr + [(Rule.new(head_aux_rule,
						   body_aux_rule)).to_s]
    end
  end
  
  ################################################################
  
  #////////////////////////////////////////////////////////////////
  # Function rewrite_body builds corresponding standard body 
  # @dlp_body of NFN body rule @dlp_body
  
  def rewrite_body(number_rule,head)
    
   num_disj = 1
   @nfn_body.each do |bdis|
      bdis.set_num_bd(num_disj)
      if bdis.size == 1 then
	# since the  basic disjunction contains only a literal
	# the aux atom is not built
	
	@dlp_body.add(bdis.to_literal)
	if !bdis[0].isNegative
	  # the atom is added to aux_hash because all its variables
	  # are safe 
	  aux = Atom.new(bdis[0].internal_atom.predicate_name,
			 bdis[0].internal_atom.terms)
	  @aux_hash = aux_hash.merge({bdis => aux})
	end
      else
	# auxiliary atom corresponding to the basic disjunction
	# is built
	aux = build_aux(bdis,number_rule, head)
	@aux_hash = @aux_hash.merge({bdis => aux})
	build_aux_rules(aux, bdis, num_disj, number_rule)
      end
      num_disj += 1
    end

    # the function builds match atoms for shared variables
    # and rename shared variables in each auxiliary atom
    atomsJoiningSharedVars(number_rule)
    addUnivSharedRules
    # the function builds match atoms for unrestricted variables
    # and rename unrestricted variables in each auxiliary atom
    atomsJoiningUnresVars(number_rule)
    
    @aux_hash.each_key do |bd|
      if bd.size > 1    
	@dlp_body.add(aux_hash[bd])
      end
    end


    @match.each do |lit|
      @dlp_body.add(lit)
    end
  end


end
